
# This file was generated by 'versioneer.py' (0.28) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2023-03-12T15:36:47-0400",
 "dirty": false,
 "error": null,
 "full-revisionid": "6b068542fac0af534840c6245c9e4dd2e191e32d",
 "version": "0.0.5"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
